// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.register.provider.aws.ddms;


import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.dynamodbv2.model.ConditionalCheckFailedException;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.provider.aws.util.DocUtil;
import org.opengroup.osdu.register.provider.interfaces.ddms.IDdmsRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

import javax.inject.Inject;
import java.util.List;

@Repository
public class AwsDdmsRepo implements IDdmsRepository {


    @Autowired
    private DpsHeaders headers;

    @Autowired
    private JaxRsDpsLog logger;

    @Inject
    private DynamoDBQueryHelperFactory dynamoDBQueryHelperFactory;

    @Value("${aws.dynamodb.ddmsTable.ssm.relativePath}")
    String ddmsTableParameterRelativePath;


    @Autowired
    DocUtil docUtil;
    private DynamoDBQueryHelperV2 getDDMSQueryHelper() {
        return dynamoDBQueryHelperFactory.getQueryHelperForPartition(headers, ddmsTableParameterRelativePath);
    }

    @Override
    public Ddms create(Ddms ddms) {
        DdmsDoc doc = DdmsDoc.mapFrom(ddms, headers.getPartitionId());
        DynamoDBQueryHelperV2 queryHelper = getDDMSQueryHelper();
        try {

            queryHelper.saveWithHashCondition(doc,DocUtil.getHashKey());
        }
        catch(ConditionalCheckFailedException e){
            logger.error(String.format("An DDMS already exists with the id: %s", ddms.getId()));
            throw new AppException(409, "Conflict", String.format("An DDMS already exists with the id: %s", ddms.getId()));
        }
        catch(AppException e)
        {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }
        return ddms;
    }

    @Override
    public Ddms get(String id) {
        DynamoDBQueryHelperV2 queryHelper = getDDMSQueryHelper();
        DdmsDoc doc = queryHelper.loadByPrimaryKey(DdmsDoc.class, id);
        if (doc == null) {
            logger.error(String.format("Ddms with id %s does not exist.", id));
            throw new AppException(404, "Not found", String.format("Ddms with id %s does not exist.", id));
        } else {
            return DdmsDoc.mapto(doc);
        }
    }

    @Override
    public List<Ddms> query(String entityType) {
        DynamoDBQueryHelperV2 queryHelper = getDDMSQueryHelper();
        String dataPartitionId = headers.getPartitionId();
        DdmsDoc gsiQuery = new DdmsDoc();
        String key = String.format("%s:%s",dataPartitionId,entityType);
        gsiQuery.setPartitionIdEntityType(key);
        return docUtil.getDdmsList(queryHelper,gsiQuery);
    }

    @Override
    public boolean delete(String id) {
        DynamoDBQueryHelperV2 queryHelper = getDDMSQueryHelper();
        try{
            DdmsDoc objectToDelete = new DdmsDoc();
            objectToDelete.setId(id);
            DynamoDBDeleteExpression deleteExpression = new DynamoDBDeleteExpression()
                    .withConditionExpression("attribute_exists(id)");
            queryHelper.deleteByObjectWithCondition(objectToDelete,deleteExpression);
        }
        catch(ConditionalCheckFailedException e){
            logger.error(String.format("Failed to delete: %s Object does not exist",id));
            throw new AppException(404, "ObjectNotFound", String.format("The Ddms you are trying to delete: %s does not exist", id));
        }
        catch(AppException e)
        {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }
        return true;

    }


}
