// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.register.provider.aws.ddms;

import com.amazonaws.services.dynamodbv2.datamodeling.*;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.*;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;

import java.sql.Timestamp;
import java.util.Set;
import java.util.stream.Collectors;

@Data
@AllArgsConstructor
@NoArgsConstructor
@Builder
@DynamoDBTable(tableName = "Register.Ddms")
public class DdmsDoc {

    @DynamoDBHashKey(attributeName = "id")
    private String id;

    @DynamoDBAttribute(attributeName = "name")
    private String name;

    @DynamoDBAttribute(attributeName = "description")
    private String description;


    @DynamoDBAttribute(attributeName = "contactEmail")
    private String contactEmail;

    @DynamoDBAttribute(attributeName = "createdDateTimeEpoch")
    private String createdDateTimeEpoch;

    @DynamoDBAttribute(attributeName = "dataPartitionId")
    private String dataPartitionId;

    @DynamoDBIndexHashKey(attributeName = "partitionIdEntityType", globalSecondaryIndexName = "entityType-index") //Added this for query api
    private String partitionIdEntityType;



    @DynamoDBTypeConverted(converter = DdmsDoc.DdmsConverter.class)
    @DynamoDBAttribute(attributeName = "interfaces")
   // private Set<RegInterfaceDoc> interfaces;
    private Set<RegisteredInterface> interfaces;



    public static class DdmsConverter implements DynamoDBTypeConverter<String, Set<RegisteredInterface>> {

        @SneakyThrows
        @Override
        public String convert(Set<RegisteredInterface> object) {
            ObjectMapper mapper = new ObjectMapper();
            return mapper.writeValueAsString(object);
        }

        @SneakyThrows
        @Override
        public Set<RegisteredInterface> unconvert(String object) {
            ObjectMapper mapper = new ObjectMapper();
            return mapper.readValue(object, new TypeReference<Set<RegisteredInterface>>() {
            });
        }
    }



    public static DdmsDoc mapFrom(Ddms ddms, String dataPartitionId) {

        DdmsDocBuilder ddmsDocBuilder = new DdmsDoc().builder()
                .id(ddms.getId())
                .name(ddms.getName())
                .description(ddms.getDescription())
                .contactEmail(ddms.getContactEmail())
                .createdDateTimeEpoch((new Timestamp(System.currentTimeMillis())).toString())
                .interfaces(ddms.getInterfaces())
                .dataPartitionId(dataPartitionId)
                .partitionIdEntityType(String.format("%s:%s",dataPartitionId,getEntityType(ddms)));

        return ddmsDocBuilder.build();
    }

    public static Ddms mapto(DdmsDoc ddmsDoc) {

        Ddms ddms = new Ddms();
        ddms.setId(ddmsDoc.getId());
        ddms.setName(ddmsDoc.getName());
        ddms.setDescription(ddmsDoc.getDescription());
        ddms.setContactEmail(ddmsDoc.getContactEmail());
        ddms.setCreatedDateTimeEpoch(Timestamp.valueOf(ddmsDoc.getCreatedDateTimeEpoch()));
        ddms.setInterfaces(ddmsDoc.getInterfaces());
        return ddms;
    }






    private  static RegInterfaceDoc getRegisteredInterfaceDoc(RegisteredInterface reginterface){
        return new RegInterfaceDoc(reginterface.getEntityType(), reginterface.getSchema());
    }

    private static RegisteredInterface getRegisteredInterface(RegInterfaceDoc doc){
        RegisteredInterface registeredInterface = new RegisteredInterface();
        registeredInterface.setEntityType(doc.getEntityType());
        registeredInterface.setSchema(doc.getSchema());
        return registeredInterface;
    }

    private static String getEntityType(Ddms ddms){

        String entityType="";
        //get the first element from the set
        RegisteredInterface ri = ddms.getInterfaces().iterator().next();
        if(ri!=null) {
            RegInterfaceDoc regIntDoc = DdmsDoc.getRegisteredInterfaceDoc(ri);
            entityType = regIntDoc.getEntityType();
        }


        return entityType;

    }

}


