// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.register.provider.aws.ddms;


import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.dynamodbv2.model.ConditionalCheckFailedException;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.runners.MockitoJUnitRunner;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperFactory;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelperV2;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;
import org.opengroup.osdu.register.provider.aws.util.DocUtil;
import org.opengroup.osdu.register.provider.aws.util.TestUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.lenient;
import static org.mockito.MockitoAnnotations.initMocks;


@RunWith(MockitoJUnitRunner.class)
public class AwsDdmsRepoTest {

    @Mock
    private DpsHeaders dpsHeaders;

    @Mock
    private DynamoDBQueryHelperV2 queryHelper;

    @Mock
    private DynamoDBQueryHelperFactory queryHelperFactory;
    @Mock
    private JaxRsDpsLog logger;

    @InjectMocks
    private AwsDdmsRepo repo;


    @Mock
    private DocUtil docUtil;


    @Before
    public void setUp() {
        lenient().doReturn(TestUtils.getDataPartitionId()).when(dpsHeaders).getPartitionId();

        initMocks(this);
        Mockito.when(queryHelperFactory.getQueryHelperForPartition(Mockito.any(DpsHeaders.class), Mockito.any()))
                .thenReturn(queryHelper);
    }

    @Test
    public void createDdms_success() {
        Ddms ddms = createMockDdms();
        Mockito.doNothing().when(queryHelper).saveWithHashCondition(Mockito.any(DdmsDoc.class), Mockito.anyString());
        Ddms resultDdms = repo.create(ddms);
        assertEquals(ddms, resultDdms);

    }

    @Test
    public void createDdms_throw409_whenIdAlreadyExists() {
        Ddms ddms = createMockDdms();
        AppException e = new AppException(409, "Conflict", String.format("A DDMS already exists with the same id: %s", ddms.getId()));
        Mockito.doThrow(e).when(queryHelper).saveWithHashCondition(Mockito.any(DdmsDoc.class), Mockito.anyString());

        try {
            repo.create(ddms);
        } catch (AppException ex) {
            assertEquals(409, ex.getError().getCode());
        }
    }

    @Test
    public void getDdms_success() {
        Ddms ddms = createMockDdms();
        String id = TestUtils.getDdms_id();
        DdmsDoc doc = DdmsDoc.mapFrom(ddms, TestUtils.getDataPartitionId());
        Mockito.when(queryHelper.loadByPrimaryKey(DdmsDoc.class, id)).thenReturn(doc);

        Ddms resultDdms = repo.get(id);
        ddms.setCreatedDateTimeEpoch(resultDdms.getCreatedDateTimeEpoch());
        assertEquals(ddms, resultDdms);

    }

    @Test
    public void getDdms_throw404_whenIdoesNotExist() {
        String id = TestUtils.getDdms_id();
        AppException e = new AppException(404, "", "");
        Mockito.doThrow(e).when(queryHelper).loadByPrimaryKey(DdmsDoc.class, id);
        try {
            repo.get(id);
        } catch (AppException ex) {
            assertEquals(404, ex.getError().getCode());

        }
    }

    @Test
    public void deleteDdms_success() {
//        Ddms ddms = createMockDdms();
        String id = TestUtils.getDdms_id();
        Mockito.doNothing().when(queryHelper).deleteByObjectWithCondition(Mockito.any(DdmsDoc.class), Mockito.any(DynamoDBDeleteExpression.class));
        Boolean result = repo.delete(id);
        assertEquals(result, true);
    }

    @Test
    public void deleteDdms_throw404_whenIdoesNotExist() {
        String id = TestUtils.getDdms_id();
        ConditionalCheckFailedException e = new ConditionalCheckFailedException("Failed to delete");
        Mockito.doThrow(e).when(queryHelper).deleteByObjectWithCondition(Mockito.any(DdmsDoc.class), Mockito.any(DynamoDBDeleteExpression.class));

        try {
            repo.delete(id);
        } catch (AppException ex) {
            assertEquals(404, ex.getError().getCode());

        }
    }

    @Test
    public void query_success() {
        Ddms ddms1 = createMockDdms();
        Ddms ddms2 = createMockDdms();
        ddms2.setId("DdmsId2");
        List<Ddms> ddmsList = new ArrayList<>();
        ddmsList.add(ddms1);
        ddmsList.add(ddms2);

        Mockito.when(docUtil.getDdmsList(Mockito.any(DynamoDBQueryHelperV2.class), Mockito.any(DdmsDoc.class))).thenReturn(ddmsList);

        List<Ddms> resultDdmsList = repo.query("type");


        ddmsList.get(0).setCreatedDateTimeEpoch(resultDdmsList.get(0).getCreatedDateTimeEpoch());
        ddmsList.get(1).setCreatedDateTimeEpoch(resultDdmsList.get(1).getCreatedDateTimeEpoch());

        assertEquals(ddmsList, resultDdmsList);

    }

    private Ddms createMockDdms() {
        Ddms ddms = new Ddms();
        RegisteredInterface ri = new RegisteredInterface();
        ri.setEntityType("type");
        ri.setSchema(Collections.singletonMap("first", "second"));
        ddms.setId(TestUtils.getDdms_id());
        ddms.setName(TestUtils.getDdms_name());
        ddms.setDescription(TestUtils.getDdms_description());
        ddms.setContactEmail(TestUtils.getAction_contactEmail());
        ddms.getInterfaces().add(ri);
        return ddms;
    }
}
