/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.action;

import com.google.gson.Gson;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.util.Crc32c;
import org.opengroup.osdu.core.gcp.osm.model.Destination;
import org.opengroup.osdu.core.gcp.osm.model.Kind;
import org.opengroup.osdu.core.gcp.osm.model.Namespace;
import org.opengroup.osdu.core.gcp.osm.model.query.GetQuery;
import org.opengroup.osdu.core.gcp.osm.service.Context;
import org.opengroup.osdu.core.gcp.osm.service.Transaction;
import org.opengroup.osdu.core.gcp.osm.translate.ExceptionClassifier;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.action.model.Filter;
import org.opengroup.osdu.register.provider.gcp.action.model.ActionDto;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.springframework.beans.BeanUtils;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import static org.opengroup.osdu.core.gcp.osm.model.where.predicate.Eq.eq;
import static org.springframework.beans.factory.config.BeanDefinition.SCOPE_SINGLETON;

@Slf4j
@Repository
@Scope(SCOPE_SINGLETON)
@ConditionalOnProperty(name = "osmDriver")
@RequiredArgsConstructor
public class OsmActionRepo implements IActionRepo {

    private static final String CONFLICT_REASON = "Conflict";

    private final Context context;
    private final TenantInfo tenantInfo;
    ReentrantLock actionDeleteLock = new ReentrantLock();

    private static final Namespace NAMESPACE = new Namespace("DE");
    private static final Kind KIND = new Kind("ACTION");

    //queries addressed to tenant specific GCP project and namespace and to a certain kind.
    private Destination getDestination() {
        return Destination.builder().partitionId(tenantInfo.getDataPartitionId()).namespace(NAMESPACE).kind(KIND).build();
    }

    @Override
    public Action createAction(Action action) {
        Transaction txn = context.beginTransaction(getDestination());
        try {
            if (context.getOne(new GetQuery<>(ActionDto.class, getDestination(), eq("id", action.getId()))) != null) {
                throwConflict();
            }
            ActionDto actionDto = getDtoByAction(action);
            context.create(actionDto, getDestination());
            txn.commitIfActive();
        } catch (AppException e) {
            throw e;
        } catch (Exception ex) {
            ExceptionClassifier.ExceptionClassification classification = context.classifyException(ex);
            switch (classification) {
                case ALREADY_EXISTS:
                    throwConflict();
                case CONFLICT:
                    throw new AppException(409, CONFLICT_REASON, "ABORTED due to conflict");
                default:
                    throw ex;
            }
        } finally {
            txn.rollbackIfActive();
        }
        return action;

    }

    @Override
    public Action get(String id) {

        ActionDto actionDto = context.getOne(new GetQuery<>(ActionDto.class, getDestination(), eq("id", id)));

        if (actionDto == null)
            throw new AppException(404, "Not found", String.format("Action with id %s does not exist.", id));

        Action action = getActionByDto(actionDto);

        if (getActionChecksum(action).compareTo(actionDto.getCheckSum()) != 0) {
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Internal server error", String.format("Error retrieving action '%s'.", actionDto.getName()));
        }
        return action;
    }

    @Override
    public boolean delete(String id) {
        boolean output = false;

        Transaction txn = context.beginTransaction(getDestination());
        try {
            actionDeleteLock.lock();
            if (context.getOne(new GetQuery<>(ActionDto.class, getDestination(), eq("id", id))) != null) {
                context.deleteById(ActionDto.class, getDestination(), id);
                txn.commitIfActive();
                output = true;
            }
        } catch (Exception ex) {
            ExceptionClassifier.ExceptionClassification classification = context.classifyException(ex);
            switch (classification) {
                case ALREADY_EXISTS:
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to delete the same DDMS.");
                case CONFLICT:
                    throw new AppException(409, CONFLICT_REASON, "ABORTED due to conflict");
                default:
                    throw ex;
            }
        } finally {
            actionDeleteLock.unlock();
            txn.rollbackIfActive();
        }
        return output;
    }


    private void throwConflict() {
        throw new AppException(409, CONFLICT_REASON, "An action already exists with the same message and endpoint combination");
    }

    @Override
    public List<Action> getAllActions() {
        List<ActionDto> actionDtos = context.getResultsAsList(new GetQuery<>(ActionDto.class, getDestination()));
        List<Action> actions = new ArrayList<>();
        for (ActionDto dto : actionDtos) {
            Action action = getActionByDto(dto);
            if (getActionChecksum(action).compareTo(dto.getCheckSum()) != 0) {
                log.error(String.format("Error retrieving action '%s'.", dto.getName()));
                throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Internal server error", "Error retrieving actions");
            }
            actions.add(action);
        }
        return actions;
    }

    String getActionChecksum(Action action) {
        String serializedAction = new Gson().toJson(action);
        return Crc32c.hashToBase64EncodedString(serializedAction);
    }

    private ActionDto getDtoByAction(Action action) {
        Filter filter = action.getFilter();

        ActionDto actionDto = new ActionDto();
        BeanUtils.copyProperties(action, actionDto);
        actionDto.setCheckSum(getActionChecksum(action));
        actionDto.setType(filter.getEntityType());
        actionDto.setSource(filter.getSource());
        actionDto.setVersion(filter.getVersion());

        return actionDto;
    }

    private Action getActionByDto(ActionDto actionDto) {
        Action action = new Action();
        BeanUtils.copyProperties(actionDto, action);

        Filter filter = new Filter();
        filter.setEntityType(actionDto.getType());
        filter.setSource(actionDto.getSource());
        filter.setVersion(actionDto.getVersion());
        action.setFilter(filter);

        return action;
    }

}
