/*
 * Copyright 2017-2020, Schlumberger
 * Copyright 2021 Google LLC
 * Copyright 2021 EPAM Systems, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.google.api.gax.rpc.NotFoundException;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriver;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriverRuntimeException;
import org.opengroup.osdu.core.gcp.oqm.model.OqmDestination;
import org.opengroup.osdu.core.gcp.oqm.model.OqmMessage;
import org.opengroup.osdu.core.gcp.oqm.model.OqmSubscription;
import org.opengroup.osdu.core.gcp.oqm.model.OqmTopic;
import org.opengroup.osdu.register.provider.gcp.mappers.oqm.OqmConfigurationProperties;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

@Slf4j
@Component
@ConditionalOnProperty(name = "oqmDriver")
@RequiredArgsConstructor
public class OqmOqmSubscription {

    private final OqmConfigurationProperties properties;
    private final OqmDriver driver;

    @PostConstruct
    void postConstruct() {
        log.info("OqmSubscription bean constructed");
    }

    public Subscription create(Subscription subscription, TenantInfo tenantInfo) throws IOException {
        createOqmSubscription(tenantInfo, subscription.getTopic(), subscription.getNotificationId());
        return subscription;
    }

    public void delete(TenantInfo tenantInfo, String subscriptionId) {
        try {
            driver.deleteSubscription(subscriptionId, getDestination(tenantInfo));
        } catch (NotFoundException e) {
            //silently catch
        }
    }

    private OqmSubscription createOqmSubscription(TenantInfo tenantInfo, String topicId, String subscriptionId) {

        OqmDestination destination = getDestination(tenantInfo);

        OqmTopic topic = driver.getTopic(topicId, destination)
                .orElseThrow(() -> new OqmDriverRuntimeException(null, "Create Subscription: Topic %s doesn't exist", topicId));

        OqmSubscription subscriptionRequest = OqmSubscription.builder().name(subscriptionId).topics(Collections.singletonList(topic)).build();

        OqmSubscription createdSubscription = driver.createAndGetSubscription(subscriptionRequest, destination);

        //Send "subscription created" event to the "subscriber control topic"
        sendSubscriptionLifecycleEvent(tenantInfo, createdSubscription);

        return createdSubscription;

    }

    /**
     * Sends "subscription created" event to the "subscriber control topic".
     * This topic is a "control channel" between Register and Notification services.
     * The former sends events on new pull Subscriptions being created, the latter listens for events and creates corresponding Subscribers.
     */
    private void sendSubscriptionLifecycleEvent(TenantInfo tenantInfo, OqmSubscription subscription) {
        Map<String, String> attributes = new HashMap<>();
        attributes.put("data-partition-id", tenantInfo.getDataPartitionId());
        attributes.put("subscription-id", subscription.getName());
        attributes.put("topic", subscription.getTopics().get(0).getName());

        OqmMessage message = OqmMessage.builder().data("Subscription Created").attributes(attributes).build();
        OqmTopic controlTopic = OqmTopic.builder().name(properties.getRegisterSubscriberControlTopicName()).build();
        try {
            driver.publish(message, controlTopic, getDestination(tenantInfo));
        } catch (Exception e) {
            log.warn("Failed attempt to send 'Subscription Created' event to the '{}' subscriber control topic", properties.getRegisterSubscriberControlTopicName());
            log.warn("* Exception thrown", e);
        }
    }

    private OqmDestination getDestination(TenantInfo tenantInfo) {
        return OqmDestination.builder().partitionId(tenantInfo.getDataPartitionId()).build();
    }
}
