/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.provider.gcp.subscriber;

import com.google.common.base.Strings;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.IKmsClient;
import org.opengroup.osdu.core.gcp.oqm.driver.OqmDriverRuntimeException;
import org.opengroup.osdu.core.gcp.osm.model.Destination;
import org.opengroup.osdu.core.gcp.osm.model.Kind;
import org.opengroup.osdu.core.gcp.osm.model.Namespace;
import org.opengroup.osdu.core.gcp.osm.model.query.GetQuery;
import org.opengroup.osdu.core.gcp.osm.service.Context;
import org.opengroup.osdu.core.gcp.osm.service.Transaction;
import org.opengroup.osdu.core.gcp.osm.translate.ExceptionClassifier;
import org.opengroup.osdu.register.provider.gcp.subscriber.model.SubscriptionDto;
import org.opengroup.osdu.register.provider.gcp.util.GcpAppServiceConfig;
import org.opengroup.osdu.register.subscriber.model.*;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.BeanUtils;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Repository;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.sql.Timestamp;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

import static org.opengroup.osdu.core.gcp.osm.model.where.predicate.Eq.eq;
import static org.springframework.beans.factory.config.BeanDefinition.SCOPE_SINGLETON;

@Repository
@Scope(SCOPE_SINGLETON)
@ConditionalOnProperty(name = "osmDriver")
@RequiredArgsConstructor
@Slf4j
public class OsmAccess implements IDatastoreAccess {

    private final Context context;
    private final TenantInfo tenantInfo;
    private final OqmOqmSubscription subscription;
    private final IKmsClient kmsClient;

    ReentrantLock subscriptionDeleteLock = new ReentrantLock();

    private static final String SERVER_ERROR_REASON = "Server Error";
    private static final String CONFLICT_REASON = "Conflict";
    private static final String ABORTED_DUE_TO_CONFLICT_MSG = "ABORTED due to conflict";

    private static final Namespace NAMESPACE = new Namespace("DE");
    private static final Kind KIND = new Kind("SUBSCRIPTION");

    //queries addressed to tenant specific GCP project and namespace and to a certain kind.
    private Destination getDestination() {
        return Destination.builder().partitionId(tenantInfo.getDataPartitionId()).namespace(NAMESPACE).kind(KIND).build();
    }

    @PostConstruct
    void postConstruct() {
        log.info("OsmAccess bean constructed");
    }

    @Override
    public Subscription create(Subscription subscriptionRequest) {
        subscriptionRequest = createSubscriptionInStorage(getDtoBySubscription(subscriptionRequest));
        try {
            subscription.create(subscriptionRequest, tenantInfo);
        } catch (Exception e) {
            deleteFromStorage(subscriptionRequest.getId(), false);
            throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error creating subscription", e);
        }
        return subscriptionRequest;
    }

    @Override
    public Subscription get(String id) {

        GetQuery<SubscriptionDto> q = new GetQuery<>(SubscriptionDto.class, getDestination(), eq("id", id));
        SubscriptionDto existingSubscription = context.getOne(q);
        if (existingSubscription == null)
            throw new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", id));

        return getSubscriptionByDto(existingSubscription);
    }

    // TODO: Remove this api when switch notification to use cache
    @Override
    public List<Subscription> query(String notificationId) {
        GetQuery<SubscriptionDto> q = new GetQuery<>(SubscriptionDto.class, getDestination(), eq("notificationId", notificationId));
        return context.getResultsAsList(q).stream().map(this::getSubscriptionByDto).collect(Collectors.toList());
    }

    @Override
    public boolean delete(String id) {
        return deleteFromStorage(id, true);
    }

    @Override
    public boolean patch(String id, Secret secret) {
        return updateSubscriptionInStorage(id, secret);
    }

    @Override
    public List<Subscription> getAll() {
        return context.getResultsAsList(new GetQuery<>(SubscriptionDto.class, getDestination()))
                .stream().map(this::getSubscriptionByDto).collect(Collectors.toList());
    }

    private Subscription createSubscriptionInStorage(SubscriptionDto subscriptionRequest) {

        Transaction txn = context.beginTransaction(getDestination());
        try {
            GetQuery<SubscriptionDto> q = new GetQuery<>(SubscriptionDto.class, getDestination(), eq("id", subscriptionRequest.getId()));
            if (context.getOne(q) != null)
                throw new AppException(409, CONFLICT_REASON, "A subscriber already exists with the same topic and endpoint combination");

            subscriptionRequest.setCreatedOnEpoch(new Timestamp(System.currentTimeMillis()));
            Subscription createdSubscription = getSubscriptionByDto(context.createAndGet(subscriptionRequest, getDestination()));
            txn.commitIfActive();
            return createdSubscription;
        } catch (AppException e) {
            throw e;
        } catch (Exception e) {
            ExceptionClassifier.ExceptionClassification classification = context.classifyException(e);
            switch (classification) {
                case ALREADY_EXISTS:
                    throw new AppException(409, CONFLICT_REASON, "A subscriber already exists with the same topic and endpoint combination");
                case CONFLICT:
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                default:
                    throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error creating subscription.", e);
            }
        } finally {
            txn.rollbackIfActive();
        }
    }

    private boolean deleteFromStorage(String id, boolean deleteSubscription) {
        boolean output = false;

        Transaction txn = context.beginTransaction(getDestination());
        try {
            subscriptionDeleteLock.lock();
            GetQuery<SubscriptionDto> q = new GetQuery<>(SubscriptionDto.class, getDestination(), eq("id", id));
            SubscriptionDto existingSubscription = context.getOne(q);
            if (existingSubscription != null) {
                if (deleteSubscription) {
                    subscription.delete(tenantInfo, existingSubscription.getNotificationId());
                }
                context.deleteById(SubscriptionDto.class, getDestination(), id);
                txn.commitIfActive();
                output = true;
            }
        } catch (Exception e) {
            ExceptionClassifier.ExceptionClassification classification = context.classifyException(e);
            switch (classification) {
                case ALREADY_EXISTS:
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to delete the same DDMS.");
                case CONFLICT:
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                default:
                    throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error deleting subscription.", e);
            }
        } finally {
            subscriptionDeleteLock.unlock();
            txn.rollbackIfActive();
        }
        return output;
    }

    private boolean updateSubscriptionInStorage(String id, Secret secret) {
        boolean output = false;

        Transaction txn = context.beginTransaction(getDestination());
        try {
            GetQuery<SubscriptionDto> q = new GetQuery<>(SubscriptionDto.class, getDestination(), eq("id", id));
            SubscriptionDto existingSubscription = context.getOne(q);
            if (existingSubscription != null) {
                Subscription updatedSubscriptionRequest = getSubscriptionByDto(existingSubscription);
                updatedSubscriptionRequest.setSecret(secret);
                context.upsert(getDtoBySubscription(updatedSubscriptionRequest), getDestination());
                txn.commitIfActive();
                output = true;
            }
        } catch (Exception e) {
            ExceptionClassifier.ExceptionClassification classification = context.classifyException(e);
            switch (classification) {
                case ALREADY_EXISTS:
                    throw new AppException(409, CONFLICT_REASON, "Another request is trying to update the same DDMS.");
                case CONFLICT:
                    throw new AppException(409, CONFLICT_REASON, ABORTED_DUE_TO_CONFLICT_MSG);
                default:
                    throw new AppException(500, SERVER_ERROR_REASON, "Unexpected error updating subscription.", e);
            }
        } finally {
            txn.rollbackIfActive();
        }
        return output;
    }

    private SubscriptionDto getDtoBySubscription(Subscription subscription) {
        SubscriptionDto subscriptionDto = new SubscriptionDto();
        BeanUtils.copyProperties(subscription, subscriptionDto);

        subscriptionDto.setSecretValue(encryptSecret(subscription.getSecret().toString()));
        subscriptionDto.setSecretType(subscription.getSecret().getSecretType());
        return subscriptionDto;
    }

    private Subscription getSubscriptionByDto(SubscriptionDto subscriptionDto) {
        Subscription subscription = new Subscription();
        BeanUtils.copyProperties(subscriptionDto, subscription);

        if (!Strings.isNullOrEmpty(subscriptionDto.getSecretValue()) && !Strings.isNullOrEmpty(subscriptionDto.getSecretType())) {

            String decryptedValue = decryptSecret(subscriptionDto.getSecretValue());
            if (subscriptionDto.getSecretType().equals(Constants.GSASecret)) {
                GsaSecret gsaSecret = new GsaSecret();
                String[] splitSecret = decryptedValue.split("`");
                gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
                subscription.setSecret(gsaSecret);
            } else {
                HmacSecret hmacSecret = new HmacSecret();
                hmacSecret.setValue(decryptedValue);
                subscription.setSecret(hmacSecret);
            }
        }
        return subscription;
    }

    public String encryptSecret(String plainString) {
        try {
            return kmsClient.encryptString(plainString);
        } catch (IOException e) {
            throw new OqmDriverRuntimeException("Attempting to encrypt Subscription secret", e);
        }
    }

    public String decryptSecret(String encryptedString) {
        try {
            return kmsClient.decryptString(encryptedString);
        } catch (IOException e) {
            throw new OqmDriverRuntimeException("Attempting to decrypt Subscription secret", e);
        }
    }
}
