/*
  Copyright 2021 Google LLC
  Copyright 2021 EPAM Systems, Inc

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */

package org.opengroup.osdu.register.provider.reference.ddms;

import static org.junit.Assert.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

import java.util.Collections;
import java.util.Optional;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.ddms.model.Ddms;
import org.opengroup.osdu.register.provider.reference.mapper.DdmsMapper;
import org.opengroup.osdu.register.provider.reference.model.document.DdmsDocument;
import org.opengroup.osdu.register.provider.reference.repository.mongo.DdmsDocumentRepository;

@RunWith(MockitoJUnitRunner.class)
public class ReferenceDdmsRepositoryTest {

  @Mock
  private DdmsDocumentRepository ddmsDocumentRepository;

  @Mock
  private DdmsMapper ddmsMapper;

  @InjectMocks
  private ReferenceDdmsRepository referenceDdmsRepository;

  private Ddms ddms;
  private DdmsDocument ddmsDocument;

  @Before
  public void setUp() {
    ddms = new Ddms();
    ddms.setId("ddmsId");

    ddmsDocument = new DdmsDocument();
    ddmsDocument.setId("documentId");
  }

  @Test
  public void should_createDdms_when_ddmsDoesNotExist() {
    when(ddmsDocumentRepository.existsById(any())).thenReturn(false);
    when(ddmsMapper.toDocument(ddms)).thenReturn(ddmsDocument);
    when(ddmsDocumentRepository.save(ddmsDocument)).thenReturn(ddmsDocument);
    when(ddmsMapper.toDdms(ddmsDocument)).thenReturn(ddms);

    assertEquals(ddms, referenceDdmsRepository.create(ddms));
  }

  @Test(expected = AppException.class)
  public void should_throwException_when_creatingDdmsWhichExists() {
    when(ddmsDocumentRepository.existsById(any())).thenReturn(true);

    referenceDdmsRepository.create(ddms);
  }

  @Test
  public void should_getDdms_when_ddmsExists() {
    when(ddmsDocumentRepository.findById(any())).thenReturn(Optional.of(ddmsDocument));
    when(ddmsMapper.toDdms(ddmsDocument)).thenReturn(ddms);

    assertEquals(ddms, referenceDdmsRepository.get(any()));
  }

  @Test(expected = AppException.class)
  public void should_throwException_when_gettingDdmsWhichDoesNotExist() {
    referenceDdmsRepository.get(any());
  }

  @Test
  public void should_deleteDdms_when_ddmsExists() {
    when(ddmsDocumentRepository.existsById(any())).thenReturn(true);

    assertEquals(true, referenceDdmsRepository.delete(any()));
  }

  @Test
  public void shouldNot_deleteDdms_when_ddmsDoesNotExist() {
    when(ddmsDocumentRepository.existsById(any())).thenReturn(false);

    assertEquals(false, referenceDdmsRepository.delete(any()));
  }

  @Test
  public void should_queryDdmsByType() {
    when(ddmsDocumentRepository.queryDmsByType(any())).thenReturn(Collections.singletonList(ddmsDocument));
    when(ddmsMapper.toDdms(ddmsDocument)).thenReturn(ddms);

    assertEquals(Collections.singletonList(ddms), referenceDdmsRepository.query(any()));
  }
}
